#!/usr/bin/env python
# coding: utf-8

from datetime import datetime, timedelta

import osr, ogr
import numpy as np
import scipy
import scipy.ndimage as ndi
import cartopy.crs as ccrs

def iter_times(start, end, step=timedelta(hours=1), exclude_end=False):
    current = start

    if exclude_end:
        while current < end:
            yield current
            current = current + step
    else:
        while current <= end:
            yield current
            current = current + step



def create_rotpole_reference(pollon, pollat):
    """\
    Return spatial reference for rotated pole coordinates.

    pollon: longitude of north pole
    pollat: latitude of north pole
    """
    ref = osr.SpatialReference()
    ref.ImportFromEPSG(4326)     # WSG-84

    if (pollon is not None) and (pollat is not None):
        o_lon_p = 0.0
        o_lat_p = pollat
        lon_0 = 180 + pollon

        string = (
            'PROJCS["rotated_pole projection",EXTENSION["PROJ4","+proj=ob_tran +o_proj=longlat'
            ' +o_lon_p=%.2f +o_lat_p=%.2f +lon_0=%.2f +a=1 +to_meter=0.0174532925199"]]'
        ) % (o_lon_p, o_lat_p, lon_0)

        ref.ImportFromWkt(string)    # rotpole transformation

    return ref






def epsg2rotpole(epsg, pollon, pollat, inverse=False):
    """\
    Create transformation from EPSG to rotpole described
    by pollon, pollat.
    """
    ref0 = osr.SpatialReference()
    ref0.ImportFromEPSG(epsg)

    ref1 = create_rotpole_reference(pollon, pollat)

    if inverse:
        return osr.CoordinateTransformation(ref1, ref0)
    else:
        return osr.CoordinateTransformation(ref0, ref1)


def wgs84_to_rotpole2(lon, lat, pollon, pollat):
    """\
    Convert points using cartopy
    """
    c = ccrs.RotatedPole(pollon, pollat)
    res = c.transform_points(ccrs.PlateCarree(), lon, lat)[:,:,:]
    rlon, rlat = res[:,:,0], res[:,:,1]

    return rlon, rlat


def wgs84_to_rotpole(lon, lat, pollon, pollat):
    """\
    Convert lon, lat to rlon, rlat.
    """
    trans = epsg2rotpole(4326, pollon, pollat)
    point = ogr.Geometry(ogr.wkbPoint)
    point.AddPoint(float(lon), float(lat))
    point.Transform(trans)
    return point.GetX(), point.GetY()




def spatial_binning(a):
    kernel = 0.25 * np.ones((1,2,2))
    return ndi.convolve(a, kernel)[:,::2,::2]


